/*GPL*START*
 * 
 * Copyright (C) 1998 by Johannes Overmann <overmann@iname.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * *GPL*END*/  

#include <math.h>
#include "qjocakediagram.h"
#include "qpainter.h"

static const QRgb colo[Qjo_cake_NUM_COLORS] = {
   0x008800, 0xff0000, 0x00ff00, 0x0000ff, 0xaa0000,
     0x888888, 0x0000aa, 0x775533, 0x005500, 0x335577};

QjoCakeDiagram::QjoCakeDiagram(QWidget *parent):QWidget(parent), format("<percent> <name> (<value>)"),
percent_format("%.1f%%"), value_format("%.2f"),
col_fg(black), col_bg(white), col_bor(black), col_cbor(black),
rot(0.0), dirty(true), cur_color(0), next_id(0), rotation(false)  {
   for(int i=0; i < Qjo_cake_NUM_COLORS; i++)
     col[i].setRgb(colo[i]);
   pix = 0;
}


QjoCakeDiagram::~QjoCakeDiagram() {
   if(pix) delete pix;
}


QPixmap QjoCakeDiagram::renderPixmap(int w, int h) {
   QPixmap p(w, h);
   int cx=0, cy=0, rad=0;
   drawcake(&p, cx, cy, rad);
   return p;
}


void QjoCakeDiagram::paintEvent(QPaintEvent *p) {
   if(dirty) {
      if(pix==0) pix = new QPixmap(rect().size());
      if(rect()!=pix->rect()) pix->resize(size());
      drawcake(pix, cx, cy, rad);
   }
   bitBlt(this, p->rect().topLeft(), pix, p->rect());
}


void QjoCakeDiagram::mousePressEvent(QMouseEvent *e) {
  //   double r = radius(cx, cy, e->pos().x(), e->pos().y());
	//   if(r>rad) return;
   double phi = argphi(cx, cy, e->pos().x(), e->pos().y());
   if(e->button()==LeftButton) { // rotate
      rotation = true;
      oldrot = phi - rot;
   } else { // notify 
      double a = phi - rot;
      while(a<0.0) a+=360.0;
      while(a>=360.0) a-=360.0;
      int i;
      for(i=0; i < slice.num(); i++) {
	 double d = slice[i].percent * 360.0;
	 a -= d;
	 if(a < 0.0) break;
      }      
      emit pressed(slice[i].name, mapToGlobal(e->pos()));
   }
}


void QjoCakeDiagram::mouseReleaseEvent(QMouseEvent *e) {
   if(e->button()==LeftButton) { // rotate
      rotation = false;
   }
}


void QjoCakeDiagram::mouseMoveEvent(QMouseEvent *e) {
   if(rotation) {
      rot = argphi(cx, cy, e->pos().x(), e->pos().y()) - oldrot;
      dirty = true;
      repaint(0,0,-1,-1,false);
   }
}


double QjoCakeDiagram::argphi(int cx, int cy, int x, int y) const {
   double xx = x - cx;
   double yy = cy - y;
   if(xx!=0.0) {
      double p = atan(yy/fabs(xx))*180.0/M_PI;
      if(xx<0.0) p=180.0-p;
      return p;
   } else {
      if(yy>=0.0) return 90.0;
      else return 270.0;
   }
}


double QjoCakeDiagram::radius(int cx, int cy, int x, int y) const {
   double xx = x - cx;
   double yy = cy - y;
   return sqrt(xx*xx+yy*yy);
}


void QjoCakeDiagram::setSlice(const TString& name, double value, const QColor& color) {
   slice[next_id].name = name;
   slice[next_id].value = value;
   slice[next_id].percent = 0.0; 
   slice[next_id].color = color;      
   name2id[name] = next_id++;
   dirty=true;
}


void QjoCakeDiagram::setSlice(const TString& name, double value) {
   setSlice(name, value, col[(cur_color++)%Qjo_cake_NUM_COLORS]);
}


void QjoCakeDiagram::setFormat(const char *format_, const char *percent_format_, const char *value_format_) {
   if(format_) format = format_;
   if(percent_format_) percent_format = percent_format_;
   if(value_format_) value_format = value_format_;
   dirty=true;
}


void QjoCakeDiagram::drawcake(QPixmap *pix, int &cx, int &cy, int& rad) {
   // setup
   pix->fill(col_bg);
   QPainter p(pix);
   int i;
   int w = pix->width();
   int h = pix->height();
   int border = 7;
   int tbar=0;
   
   if((w>=20)&&(h>=20)) {
      if(title) {
	 tbar = p.boundingRect(0, 5, w, h-5, AlignHCenter|AlignTop, title).height()+5;
      }
      QRect cake(border, border + tbar, w-2*border, h-2*border-tbar);
      cx = w/2;
      cy = cake.top() + cake.height()/2;
      rad = w/2-border;
      
      // calc percent
      double sum = 0.0;
      for(i=0; i < slice.num(); i++) 
	sum += slice[i].value;
      if(sum==0.0) sum=0.1;
      for(i=0; i < slice.num(); i++) 
	slice[i].percent = slice[i].value / sum;
      
   
      // draw slices
      double a = rot;
      p.setPen(QPen(col_cbor, 1));
		p.setBrush(black);
		p.drawEllipse(cake);
      for(i=0; i < slice.num(); i++) {
	 while(a<0.0) a+=360.0;
	 while(a>=360.0) a-=360.0;
	 double d = slice[i].percent * 360.0;
	 p.setBrush(slice[i].color);
	 p.drawPie(cake, int(16.0*a), int(16.0*d));
	 a += d;
      }
      
      // draw labels
      a = rot;
      p.setBrush(NoBrush);
      p.setFont(QFont("helvetica", 12, QFont::Bold));      
      p.setPen(col_fg);
      for(i=0; i < slice.num(); i++) {
	 double d = slice[i].percent * 360.0;
	 double d2 = (a+d/2)*M_PI/180.0;
	 double rx = double(cake.width())/3.5;
	 double ry = double(cake.height())/3.5;
	 int x = int(rx*cos(d2));
	 int y = int(-ry*sin(d2));
	 p.drawText(x, y+tbar, w, h-tbar, AlignHCenter|AlignVCenter, label(slice[i]));
	 a += d;
      }
      
      // draw title
      if(title) {
	 p.setPen(col_fg);      
	 p.drawText(0, 5, w, h-5, AlignHCenter|AlignTop, title);
      }
      
   }
   
   // draw border
   p.setPen(col_bor);
   p.drawRect(pix->rect());
   
   dirty = false;
}


TString QjoCakeDiagram::label(const Slice& sl) const {
   TString l(format);
   char vbuf[1024];
   char pbuf[1024];
   
   sprintf(vbuf, value_format, sl.value);
   sprintf(pbuf, percent_format, sl.percent*100.0);
   l.searchReplace("<name>", sl.name);
   l.searchReplace("<value>", vbuf);
   l.searchReplace("<percent>", pbuf);
   return l;
}


void QjoCakeDiagram::clearSlices() {   
   name2id.empty();
   slice.empty();
   next_id=0;
   cur_color=0;
   dirty = true;
}


